/**************************************************************

   Program Title:  P1 AVR Boot/Monitor
   
   Program Name:  P1_AVR_Monitor.c	  Version: 0.91
   
   Author:  SS	  					  Date:  October 10, 2004
   
   Compiled with ImageCraft ICCAVR Standard V7.02 
   
   Fuses:  FF D9 E2
   
   Version for ATmega644.
   
   The P1 AVR Boot Monitor is a set of functions that configure the 
   AVR running on the P1 processor for general operation and communication 
   with the P1 Blackfin processor.  Not all of these routines are required 
   for P1 operation but have been included to simplify custom application 
   development.   The AVR is a smart, programmable, IO engine for the Blackfin
   with TWI, analog and digital IO that can be monitored and controlled by 
   user applications running independently on the AVR.  
   
   The AVR Boot Monitor resets the Blackfin to boot mode 3 at power up allowing
   the Blackfin to boot from external serial flash.  If this code is modified
   the Blackfin control routines should not be modified or the Blackfin may
   be left in a reset state.  The Blackfin and AVR share the SPI lines so 
   any use of the AVR SPI functions should be coordinated with Blackfin use.
   The serial flash can be programmable by the AVR.  In this release the AVR 
   leaves the SPI port in default slave mode to support communication with 
   the Blackfin. 
   
   Revn 0.87 and above support all Serial Flash devices.  Note that the 
   AT45DB321D and AT45DB642D are automatically programmed to power of 2
   operation before programming.  The board must be re-powered before loading
   code into the Flash.
   
   Key functions:
   	   Blackfin Boot Control
	   Blackfin Serial Flash programming/access
	   (Partial) TWI 2-Wire Master Interrupt Controller
	   Real Time Clock setup with four configurable software timers
	   Full duplex interrupt drive USART with optional XON/XOFF flow control
	   1msec timer with four configurable software timers
	   Analog controller setup/access
	   TBD - Digital port configuration
	   IntelHex32 download/upload for EEPROM, AVRFLASH and Serial Flash
	   TBD - RS-485 software UART configuration
	   Auto internal clock calibration
	   Dynamic baud rate adjustment
	   Setup parameters loaded from EEPROM
	   
   Standard printf and scanf functions are not used in this version 
   due to code size constraints.  The P1 is available with an ATmega32
   with twice the Flash and SRAM if printf and scanf functionality is 
   important or a custom application requires more than the available
   AVR Flash.  ATmega164/324/624 are also pin compatible with 16/32/64K
   of Flash - the 164/324 have two UARTS - the 642 one.
   
   Copyright 2004-2007, SOC Robotics, Inc.  All rights reserved.
   
****************************************************************/
/*
 * Copyright (C) 2005-2007 by SOC Robotics, Inc. All rights reserved.
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions
 * are met:
 *
 * 1. Redistributions of source code must retain the above copyright
 *    notice, this list of conditions and the following disclaimer.
 * 2. Redistributions in binary form must reproduce the above copyright
 *    notice, this list of conditions and the following disclaimer in the
 *    documentation and/or other materials provided with the distribution.
 * 3. Neither the name of the copyright holders nor the names of
 *    contributors may be used to endorse or promote products derived
 *    from this software without specific prior written permission.
 * 4. Redistribution is only allowed in combination with SOC Machines, 
 *    Inc. hardware products.
 *
 * THIS SOFTWARE IS PROVIDED BY SOC ROBOTICS, INC. AND CONTRIBUTORS
 * ``AS IS'' AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT
 * LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS
 * FOR A PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL OC MACHINES, 
 * INC. OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT,
 * INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING,
 * BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS
 * OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED
 * AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY,
 * OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF
 * THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF
 * SUCH DAMAGE.
 *
 * For additional information see http://www.soc-machines.com
 */

/*
 * $Log: P1_Monitor.c,v $
 * Revision 0.91  2007/07/23 21:40:20  SS
 *    Port to ATmega644
 * Revision 0.90  2006/07/17 16:40:20  SS
 *    Automatic Serial Flash support to 8Mx8, Power of 2 support
 * Revision 0.87  2005/07/17 16:40:20  SS
 *    Intel Hex 2 Record support, drop OK, osc calibration.
 * Revision 0.85  2005/07/17 16:40:20  SS
 *    2-wire interface, 8MHz clock, AVR Flash writer, 9600baud
 *    real time clock.
 * Revision 0.52  2005/06/04 18:36:42  SS
 * Preview release
 * 
 */
#include <iom644v.h>
#include <macros.h>
#include <string.h>
#include <stdlib.h>
#include <stdio.h>
#include "P1_AVR_Uart.h"
#include "P1_AVR_Defines.h"
#include "P1_AVR_Extern.h"

//  Variable declarations
unsigned int bytesonline,inteladdress,intelextendedaddress;
long  lastaddress;
unsigned char programmode,crc,intelbinary[48];  // length can vary
unsigned char hexbyte[3],hexaddress[5],intelrecordcommand;
unsigned char seconds,minutes,hours,days,rtcclockcnt;
unsigned char flashtype0,flashtype1,flashpageloaded,flashpageboundarytraversed;
unsigned char flashprogrammingactive,flashdatainbuffer;
unsigned char xofftransmitted,showascii;
int	flashpagesize0,flashpagesize1,startaddress,currrentaddress;
int oldpage,oldbyteoffset,totalrxcnt;
int rtccountdown1,rtccountdown2,rtccountdown3,rtccountdown4;
int mseccountdown1,mseccountdown2,mseccountdown3,flashcountdown;

extern int _textmode;

/* Static Variables */
unsigned char UART_RxBuf[UART_RX_BUFFER_SIZE];
unsigned char UART_RxHead;
unsigned char UART_RxTail;
unsigned char UART_TxBuf[UART_TX_BUFFER_SIZE];
unsigned char UART_TxHead;
unsigned char UART_TxTail;
char  xoffsent, xonsent, enablexonoff, echoinchars, sendxonxoff;
char  rxcnt,checkflag,rxcntmax;
unsigned char  ip[5],ipmask[5],ipgate[5],instrg[8];

unsigned char FlashAddress[3];
const char signon1[] = "\n\r\n\rP1 AVR Monitor V0.90\n\r";
const char signon2[] = "Copyright 2004-2006,  SOC Robotics, Inc.\n\r\n\r";
const char signon3[] = "   D[E|A|F]  - Display AVR EEPROM, AVR Flash or Flash memory\n\r";
const char signon4[] = "   P[E|A|F]  - Program AVR EERPOM, AVR Flash or Flash memory\n\r";
const char signon5[] = "   R   - Reset Blackfin and release\n\r";
const char signon6[] = "   H   - Halt/Release Blackfin (Red LED on - Blackfin Halted)\n\r";
const char signon63[] = "   M   - Set Blackfin Bootmode [0,2,3] 3-default\n\r";
const char signon65[] = "   B   - Change baud rate\n\r";
const char signon66[] = "\n\r Enter Blackfin Bootmode [0,2,3]: ";
const char signon67[] = "\n\r Blackfin Bootmode set to ";
const char signon7[] = "   V   - Version\n\r";
const char signon8[] = "   G   - Test AVR\n\r";
const char signon82[] = "   C   - Calibrate Internal Oscillator\n\r";
const char signon83[] = "   T   - Get/Set Time\n\r";
const char signon85[] = "   A   - Read all analog channels\n\r";
const char signon9[] = "   S   - Signon message\n\r";
const char signon10[] = "  Note:  Blackfin held in reset until released by R command\n\r\n\r-";
const char signon11[] = "   Serial flash 0 ";
const char signon12[] = "   Serial flash 1 ";
const char signon121[] = "AT45DB041";
const char signon122[] = "AT45DB081";
const char signon123[] = "AT45DB161";
const char signon124[] = "AT45DB321";
const char signon125[] = "AT45DB642";
const char signon126[] = "No Flash";
const char crlfbar[] = "\n\r-";

const char displaymessage1[] = "Enter start address and length: ";
const char displaymessage2[] = "\n\rDisplay AVR Flash from 0x0054 to 254\n\r";
const char displaymessage3[] = "\n\rDisplay SPI Flash from 0 to 100\n\r";
const char displaymessage33[] = "\n\rNo mode selected\n\r";

const char progmessage[] = "P1 AVR Monitor V";

const char displaymessage4[] = "\n\rLoad EEPROM from address ";
const char displaymessage5[] = " length ";
const char displaymessage6[] = "\n\rLoad SPI Flash from address ";

const char loadmessage1[] = "\n\rLoad EEPROM\n\r";
const char loadmessage2[] = "\n\rLoad SPI Flash\n\r";
const char loadmessage3[] = "\n\rLoad AVR Flash\n\r";

const char calmessage1[] = "Calibrate Internal Osc: old value=";
const char calmessage2[] = "New Calibration value=";
const char calmessage3[] = "Store in EEPROM? [y|n]: ";
const char changebaudmessage[] = "Enter new baud rate [2400,4800,9600,19200,38400,57600]: ";

// Function declarations
unsigned char EEPROMwrite(int,unsigned char);
unsigned char EEPROMread(int);
int printf(CONST char *, ...);
int sprintf(char *, CONST char *, ...);
int scanf(CONST char *, ...);
int sscanf(char *, CONST char *, ...);

//*************************************************************
//  Main work starts here
//*************************************************************
void main()
{
 	char toggle,slaverx;
	unsigned char inchar,inbyte,twistatus1,twistatus2,twistatus3,twistatus4;
	unsigned char inbytes[5];
	int i,icnt,j,loopsetting,loopmode,blackfinresetmode,countervalue,numbr;
	
	_textmode = 0;  // do not expand /n to CRLF
	Init_AVR();
	rxcntmax = totalrxcnt = 0;
	
	loopsetting = 1000;   // 1 second delay
	loopmode = 0;  toggle = 0;
	mseccountdown1 = loopsetting;

	while(0) {
	    BlinkRed(2);
	    BlinkGreen(2);
		}
	
	BlinkRed(2);
	BlinkGreen(2);
	
	LoadEEPROMSettings();
	
	// Get Serial Flash status
	CheckFlashTypes();
	
	// Reset and release Blackfin
	ResetBlackfinBootMode(BOOTMODE3);
	ReleaseBlackfin(QUIET);	
	
	// Show something is happening
	BlinkRed(2);
	BlinkGreen(2);
	signon();
	
	// Debug - clean out the PC serial driver
	// Remove when serial driver reliable
//	for(i=0; i<10; i++)  putchar(XON);
	
	blackfinresetmode = 0x00;

	// Wait for commands on serial channel
	while (1) {
		  // Work starts here
		  if((inchar = getcharnw())) {
			   	switch (inchar) {
					 case 's': case 'S':
						 signon();
						 //puts("rxcnt="); printint(rxcnt,3,' ');
						 //if(xofftransmitted==YES)  puts(" xoff sent");
						 //else puts("xoff not sent"); 
						 //putCRLF(); putchar('-');
						 break;
					    case 'd': case 'D':
						    putchar(inchar);
						    DisplayMemory();
							putchar('-');
							break;
					    case 'p': case 'P':
							putchar(inchar);
						    LoadMemory();
							putchar('-');
							break;
					    case 'a': case 'A':
							putchar(inchar);
							putCRLF();
						    DumpAnalogChs();
							putchar('-');
							break;
					    case 'v': case 'V':
						    putchar(inchar);
							putCRLF();
						    ProgramVersion();
							putchar('-');
							break;
					    case 'c': case 'C':
						    putchar(inchar);
							putCRLF();
						    CalibrateOsc(VERBOSE);
							putchar('-');
							break;
					    case 'r': case 'R':
							Release_SPI();
						    ResetBlackfin(VERBOSE);
							putchar('-');
							break;
					    case 'm': case 'M':
							putchar(inchar);
							SetBlackfinBootMode();
							putchar('-');
							break;
					    case 'b': case 'B':
							putchar(inchar);
							EnterNewBaudRate();
							//putchar('-');
							break;
					    case 'h': case 'H':
							if(blackfinresetmode==0x01) {
								 ReleaseBlackfin(VERBOSE);
								 blackfinresetmode = 0x00;
								 }
							else {
								 HaltBlackfin(VERBOSE);
								 blackfinresetmode = 0x01;
								 }
							putchar('-');
							break;
					    case ':':
						    ProgramMemory();
							//putchar('-');
							break;
						case 'i': case 'I':
						    InternetSettings();
							putchar('-');
							break;
						case '9':
						    puts("9\n\rRelease PC5, PC4 and PB4\n\r-");
						    DDRC &= 0xCF;  // PF0
							PORTC != 0x20;  // Activate pull up
							DDRB &= 0xEF;  // Connected to PF0 via 1K
							break;
						case '8':
						    puts("8\n\rRead Flash Status Continuously");
						    Init_SPI_Master();
							ResetBlackfinBootMode(BOOTMODE0);
							//while(!(inchar=getcharnw())) { 	
								//SPI_GetManufacturerID(FLASHCHIP_1);
								SPI_SerialFlashStatus(FLASHCHIP_1);
								//shortdelay(100);
								//}
							Release_SPI();
							ResetBlackfinBootMode(BOOTMODE3);
							putCRLF();
							putchar('-');
							break;
						case '7':
						    puts("7\n\rRead Flash Manufacturer ID\n\r");
						    Init_SPI_Master();
							ResetBlackfinBootMode(BOOTMODE0);	
							SPI_GetManufacturerID(FLASHCHIP_1);
							Release_SPI();
							ResetBlackfinBootMode(BOOTMODE3);
							putCRLF();
							putchar('-');
							break;
						case '6':
						    puts("6\n\rCheck Flash Type\n\r");
						    CheckFlashTypes();
							putchar('-');
							break;
						case 'e': case 'E':
						    EraseSerialFlash();
							puts("Erase\n\r");
							putchar('-');
							break;
						case 'l': case 'L':
						    while(!(inchar=getcharnw())) putchar(0x55);
							putCRLF();
							putchar('-');
							break;
						case 'g': case 'G':
						    while(!(inchar=getcharnw()))  TestAVR();
							signon();
							break;
						case 't': case 'T':
						    putchar(inchar);
							putCRLF();
						    getsettime();
							putCRLF();
							putchar('-');
							break;
						case 'x': case 'X':
							putchar(inchar);
							TWI_Write_Slave();
							putchar('-');
							break;
						case LF:
							putCRLF();
							putchar('-');	
							break;
						case CR:
							putCRLF();
							putchar('-');	
							break;
						default:
							putchar(inchar);
							//Delaymsec(80);
							break;
						}
			   		
			   		} 

			   // Blink green led slowly to show AVR is alive
			   if(mseccountdown1==0) {
			   		mseccountdown1 = loopsetting;
					if(loopmode)  {
					    GREEN_LED(ON);
						loopmode = 0;
						}
					else {
						GREEN_LED(OFF);
						loopmode = 1;
						}
			   	 	}
					
				// Clean up flash state machine if no new Intel
				// Hex records for 3 seconds
				if((flashcountdown==0)&&(flashprogrammingactive==YES))
				        CleanUpFlashState();
		}
}
